@@ -20,7 +20,7 @@ package executable_seq
 import (
 	"context"
 	"errors"
-	"github.com/IrineSistiana/mosdns/v2/dispatcher/handler"
+	"github.com/IrineSistiana/mosdns/v3/dispatcher/handler"
 	"github.com/miekg/dns"
 	"testing"
 )
@@ -72,8 +72,10 @@ func Test_ParallelNode(t *testing.T) {
 				WantR:     tt.r2,
 				WantErr:   tt.e2,
 			}
-			handler.MustRegPlugin(p1, false)
-			handler.MustRegPlugin(p2, false)
+
+			handler.PurgePluginRegister()
+			handler.MustRegPlugin(p1)
+			handler.MustRegPlugin(p2)
 
 			qCtx := handler.NewContext(new(dns.Msg), nil)
 			err := handler.ExecChainNode(ctx, qCtx, handler.WrapExecutable(parallelNode))
