@@ -19,8 +19,8 @@ package sleep
 
 import (
 	"context"
-	"github.com/IrineSistiana/mosdns/v2/dispatcher/handler"
-	"github.com/IrineSistiana/mosdns/v2/dispatcher/pkg/pool"
+	"github.com/IrineSistiana/mosdns/v3/dispatcher/handler"
+	"github.com/IrineSistiana/mosdns/v3/dispatcher/pkg/pool"
 	"time"
 )
 
@@ -36,7 +36,7 @@ func init() {
 	handler.MustRegPlugin(&sleep{
 		BP: handler.NewBP("_sleep_500ms", PluginType),
 		d:  time.Millisecond * 500,
-	}, true)
+	})
 }
 
 // Args is the arguments of plugin. It will be decoded from yaml.
